#!/usr/bin/python
#
# RHN Registration Client
# Copyright (c) 2000, 2001 Red Hat, Inc.
#
# Author: Preston Brown <pbrown@redhat.com>

import xmlrpclib, cgiwrap
import rpm
import os, sys
import socket
import string
import string
import time

from translate import _, N_, cat

try:
    from rhn import rpclib
except ImportError:
    rpclib = __import__("xmlrpclib")

# global variables
#SYSID_DIR = /tmp
SYSID_DIR = "/etc/sysconfig/rhn"
SYSID_FILE = "%s/systemid" % SYSID_DIR
SYSID_BU_FILE = "%s/systemid.save" % SYSID_DIR


# Where do we keep the CA certificate for RHNS?
# The servers we're talking to need to have their certs
# signed by one of these CA.
rhns_ca_cert = "/usr/share/rhn/RHNS-CA-CERT"


import config
cfg = config.RHNConfig()
cfg.load()

def version():
    # substituted to the real version by the Makefile at installation time.
    return "2.7.9"

# reload the config in case it's changed on disk
def reloadConfig():
	cfg.load()

def hasSSL():
    try:
        from httplib import HTTPS
    except ImportError:
        return 0

    return 1

serverUrl = ""
try:
    from __server import serverUrl
except ImportError:
    if hasSSL():
        serverUrl = cfg.readEntry("serverURL")
    else:
        serverUrl = cfg.readEntry("noSSLServerURL")


def getProxySetting():
    proxy = None
    proxyHost = cfg.readEntry("httpProxy")
    # legacy for backwards compat
    if proxyHost == "":
        try:
            proxyHost = cfg.readEntry("pkgProxy")
        except:
            proxyHost = None

    if proxyHost:
        if proxyHost[:7] == "http://":
            proxy = proxyHost[7:]
        else:
            proxy = proxyHost

    return proxy

class Error:
    # base class for client errors
    def __init__(self, errmsg):
        self.errmsg = errmsg

    def __repr__(self):
        return self.errmsg
    
class RpmError(Error):
    """rpm itself raised an error condition"""
    def __repr__(self):
        msg = _("RPM error.  The message was:\n") + self.errmsg
        return msg

class FileError(Error):
    def __init__(self, errmsg):
        Error.__init__(self, errmsg)

    def __repr__(self):
        return _("Disk error.  The message was:\n") + self.errmsg

class DelayError(Error):
    def __init__(self, errmsg):
        Error.__init__(self,"Delay requested from server.")
        self.time = string.atoi(errmsg)

    def __repr__(self):
        return _("Server said come back in %s minutes.") % self.time

    
class CommunicationError(Error):
    def __init__(self, errmsg):
        Error.__init__(self, errmsg)

    # indicates a problem doing xmlrpc communication
    def __repr__(self):
        return _("Error communicating with server.  The message was:\n") + self.errmsg

class ValidationError(Error):
    def __init__(self, errmsg):
        Error.__init__(self, errmsg)

    # indicates an error during server input validation
    def __repr__(self):
        return _("Error validating data at server:\n") + self.errmsg

class OemInfoFileError(Error):
    def __init__(self,errmsg):
        Error.__init__(self, errmsg)

    def __repr__(self):
        return _("Error parsing the oemInfo file at field:\n") + self.errmsg

def doCall(method, *args):
    """wrapper funtion for xmlrpc calls so we can retry on network failures"""
    # note, be careful with method here, as it represents a rpc, so doing about
    # anything to it causes weird network things, ie dir(method) etc...
    ret = None

    attempt_count = 1
    if cfg.readEntry("networkRetries"):
        attempts = cfg.readEntry("networkRetries")
    else:
        attempts = 5
    
    while 1:
        failure = 0
        ret = None
        try:
            ret = apply(method, args)
        except KeyboardInterrupt:
            raise CommunicationError(_("Connection aborted by the user"))
        # if we get a socket error, keep trying
        except (socket.error, socket.sslerror), e:
            sys.stderr.write("A socket error occured: %s, attempt #%s\n" % (e, attempt_count))
            if attempt_count >= attempts:
                if len(e.args) > 1:
                    raise CommunicationError(e.args[1])
                else:
                    raise CommunicationError(e.args[0])
            else:
                failure = 1
        except cgiwrap.ProtocolError, e:
            sys.stderr.write("A protocol error occured: %s , attempt #%s\n" % (e.errmsg, attempt_count))
            if attempt_count >= attempts:
                # probabaly a better code
                if e.errcode == 404:
                    raise CommunicationError(e.errmsg)
                if attempt_count >= attempts:
                    raise CommunicationError(e.errmsg)
            else:
                failure = 1
        except rpclib.ResponseError:
            raise CommunicationError("Broken response from the server.");

        if failure:
            # rest for a second before trying again
            time.sleep(1)
            attempt_count = attempt_count + 1
        if ret != None:
            break
        
    return ret
    

def validateEmail(email):
    ret = 1

    if len(email) < 6:
        ret = 0
    if not string.find(email, "@") > 0:
        ret = 0
    if not string.find(email, ".") > 0:
        ret = 0

    return ret

def getArch():
    s = os.uname()[4]
    #s = re.sub("i.86", "i386", s)
    return s

def getOSVersionAndRelease():
    try:
        db = rpm.opendb(0, "/")
    except rpm.error:
        print "Unable to open rpm database. Perhaps another application has "+\
              "it open?"
        sys.exit()
    try:
        index = db.findbyprovides("redhat-release")[0]
    except:
        raise RpmError("Couldn't determine what version of Red Hat Linux you "
                       "are running.\nIf you get this error, try running \n\n"
                       "\t\trpm --rebuilddb\n\n")
    version = (db[index])['version']
    release = (db[index])['name']
    return (release, version)

def getVersion():
    release, version = getOSVersionAndRelease()
    return version

def getOSRelease():
    release, version = getOSVersionAndRelease()
    return release
	

# product info structure
productInfo = {
    "title" : "",
    "first_name" : "",
    "last_name" : "",
    "company" : "",
    "position" : "",
    "address1" : "",
    "address2" : "",
    "city" : "",
    "state" : "",
    "zip" : "",
    "country" : "",
    "phone" : "",
    "fax" : "",
    "contact_email" : 1,
    "contact_mail" : 0,
    "contact_phone" : 0,
    "contact_fax" : 0, 
    "newsletter" : 0,
    "special_offers" : 0
    }


def getOemInfo(info={}):
    if cfg.readEntry("oemInfoFile"):
        configFile = cfg.readEntry("oemInfoFile")
    else:
        configFile = "/etc/sysconfig/rhn/oeminfo"

    if not os.access(configFile, os.R_OK):
        return {}

    fd = open(configFile, "r")
    L = fd.readlines()

    info = {}
    for i in L:
        i = string.strip(i)
        if i == "":
            continue
        try:
            (key, value) = string.split(i, ':')
        except ValueError:
            raise OemInfoFileError(i)
        
        info[key] = string.strip(value)

    return info
    
    
def buildPackageList(callBack = None):
    list = []
    count = 1
    total = 0
    
    try:
        db = rpm.opendb(0, "/")
    except rpm.error:
        print "Unable to open rpm db. Perhaps it is in use by another application?"
        sys.exit()
        
    if "firstkey" in dir(db):
        key = db.firstkey()
        h = db[key]
    else:
        iterator = db.match(0)
        h = iterator.next()
        
    while (h):
        count = count + 1
        if "nextkey" in dir(db):
            key = db.nextkey(key)
            if key:
                h = db[key]
            else:
                h = None
            
        else:
            h = iterator.next()
        
    total = count
    count = 1
    if "firstkey" in dir(db):
        key = db.firstkey()
        h = db[key]
    else:
        iterator = db.match(0)
        h = iterator.next()
    while (h):
        name = h['name']
        epoch = h['epoch']
        if epoch == None:
            epoch = ""
        version = h['version']
        release = h['release']

        list.append([name, version, release, epoch])
        if not callBack == None:
            callBack(count, total)
        count = count + 1
        if "nextkey" in dir(db):
            key = db.nextkey(key)
            if key:
                h = db[key]
            else:
                h = None
        else:
            h = iterator.next()

    list.sort()
    return list
        
def registered():
    global SYSID_FILE
    return os.access(SYSID_FILE, os.R_OK)

def writeSystemId(systemId):
    global SYSID_FILE
    global SYSID_DIR
    
    if not os.access(SYSID_DIR, os.W_OK):
        try:
            os.mkdir(SYSID_DIR)
        except:
            return 0

    if not os.access(SYSID_DIR, os.W_OK):
        return 0

    if os.access(SYSID_FILE, os.F_OK):
	# already have systemid file there; let's back it up
	try:
	    os.rename(SYSID_FILE, SYSID_BU_FILE)
	except:
	    return 0

    f = open(SYSID_FILE, "w")
    f.write(systemId)
    f.close()

    try:
        os.chmod(SYSID_FILE, 0600)
    except:
        return 0
    return 1

def getServer(refreshCallback=None):

    if cfg.readEntry("enableProxy"):
        proxyHost = getProxySetting()
    else:
        proxyHost = None

    if hasSSL():
        serverUrl = cfg.readEntry("serverURL")
    else:
        serverUrl = cfg.readEntry("noSSLServerURL")

    if cfg.readEntry("enableProxyAuth"):
        if cfg.readEntry("proxyUser"):
            proxyUser = cfg.readEntry("proxyUser")
	else:
	    proxyUser = ""         
  
        if cfg.readEntry("proxyPassword"):
            proxyPassword = cfg.readEntry("proxyPassword")
	else:
	    proxyPassword = ""

    else:
        proxyUser = None
        proxyPassword = None

    lang = None
    for env in 'LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG':
        if os.environ.has_key(env):
            if not os.environ[env]:
                # sometimes unset
                continue
            lang = string.split(os.environ[env], ':')[0]
            lang = string.split(lang, '.')[0]
            break

        
    s = rpclib.Server(serverUrl, proxyHost,
                         refreshCallback=refreshCallback,
                         username=proxyUser,
                         password=proxyPassword)
    if lang:
        s.setlang(lang)

    # force the validation of the SSL cert
    s.use_CA_chain(rhns_ca_cert)    

    return s

def welcomeText():
    s = getServer()

    try:
        return doCall(s.registration.welcome_message)
    except rpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)
    

def privacyText():
    s = getServer()

    try:
        return doCall(s.registration.privacy_statement)
    except rpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)
    

def reserveUser(username, password):
    s = getServer()
    
    try:
        ret = doCall(s.registration.reserve_user, username, password)
    except rpclib.Fault, f:
        if f.faultCode == -3:
            # account already in use
            raise ValidationError(f.faultString)
        elif f.faultCode == -14:
            # too short password
            raise ValidationError(f.faultString)
        elif f.faultCode == -15:
            # bad chars in username
            raise ValidationError(f.faultString)
        elif f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)

    return ret

def validateRegNum(regNum):
    s = getServer()

    try:
        doCall(s.registration.validate_reg_num,regNum)
    except rpclib.Fault, f:
        if f.faultCode == -16:
            # invalid
            raise ValidationError(f.faultString)
        elif f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)


def registerUser(username, password, email = None, orgid = None, orgpassword = None):
    s = getServer()

    try:
        if not email == None:
            if orgid and orgpassword:
                doCall(s.registration.new_user, username, password, email, orgid, orgpassword)
            else:
                doCall(s.registration.new_user, username, password, email)
        else:
                doCall(s.registration.new_user, username, password)
    except rpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)


def registerSystem(username = None, password = None,
                   profileName = None, packages = None, token = None):
    s = getServer()

    auth_dict =  { "profile_name" : profileName,
                   "os_release" : getVersion(),
		   "release_name" : getOSRelease(),
                   "architecture" : getArch() };


    if token:
        auth_dict["token"] = token
    else:
        auth_dict["username"] = username
        auth_dict["password"] = password

    try:
        if packages == None:
            ret = doCall(s.registration.new_system,
                         auth_dict)
        else:
            ret = doCall(s.registration.new_system,
                         auth_dict,
                         packages)
    except rpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)

    return ret


def registerProduct(systemId, productInfo, oemInfo={}):
    s = getServer()

    try:
        doCall(s.registration.register_product, systemId, productInfo)
    except rpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)


def sendSerialNumber(systemId, num):
    s = getServer()

    try:
        if cfg.readEntry("oemId") != None:
            doCall(s.registration.send_serial, systemId, num,
                                       cfg.readEntry("oemId"))
        else:
            doCall(s.registration.send_serial, systemId, num)
    except rpclib.Fault, f:
        print f
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)


def sendHardware(systemId, hardwareList):
    s = getServer()

    try:
        doCall(s.registration.add_hw_profile, systemId, hardwareList)
    except rpclib.Fault, f:
        print f
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)

def sendPackages(systemId, packageList):
    s = getServer()

    try:
        doCall(s.registration.add_packages, systemId, packageList)
    except rpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)

def listPackages(systemId):
    s = getServer()

    try:
        print doCall(s.registration.list_packages,systemId)
    except rpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)


def updatePackages(systemId):
    s = getServer()

    try:
        doCall(s.registration.add_packages, systemId, buildPackageList())
    except rpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)
